////////////////////////////////////////////////////////////////////////////////
//	Copyright (c) CitizenLab, 2006. All Rights Reserved.
//	The latest version of this code is available at http://psiphon.civisec.org/
//
//	This software is open source; you can redistribute it and/or modify it 
//	under the terms of the GNU General Public License as published by the 
//	Free Software Foundation; either version 2 of the License, or (at your 
//	option) any later version.
//
//	This program is distributed WITHOUT ANY WARRANTY; without even the 
//	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//	See the GNU General Public License for more details at:
//	http://psiphon.civisec.org/
//	
//	This General Public License does NOT permit incorporating this software 
//	into proprietary programs.
////////////////////////////////////////////////////////////////////////////////

#include "convert.h"
#include "CppSQLite3.h"
#include "User.h"
#include "SHA1.h"
#include "strfunc.h"

#define _T(X) X

void escape(char* pDest, int nSize, const char* pSrc);

char* CDBUser::m_sDB = _T("users.db");

LOG_ROUTINE CDBUser::m_pLogFunc = NULL;

CDBUser::CDBUser()
{
	Init();
}

CDBUser::~CDBUser()
{
	Cleanup();
}

void CDBUser::Init()
{
	m_sUserName = NULL;
	m_sPassword = NULL;
	m_sEmail = NULL;
	m_sFullName = NULL;

	m_nID = 0;
	m_nDeleted = false;

	m_nLastAccess = 0;
	m_nWritten = 0;
	m_nRead = 0;
	m_nWriteTime = 0;
	m_nReadTime = 0;
}

void CDBUser::Cleanup()
{
	delete m_sUserName;
	delete m_sPassword;
	delete m_sEmail;
	delete m_sFullName;

	Init();
}

int CDBUser::Get(int nUserID)
{
	if(nUserID <= 0)
		return -1;

	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
			return false;

		CppSQLite3Statement stmt = db.compileStatement(_T("select nIndex, sUsername, sPassword, sEmail, sFullname from users where nIndex = ?;"));

		stmt.bind(1, nUserID);

		CppSQLite3Query q = stmt.execQuery();

		if(!q.eof())
		{
			SetID(q.getIntField(0));

			CConverter wUsername(q.fieldValue(1));
			SetUsername(wUsername);

			CConverter wPassword(q.fieldValue(2));
			SetPassword(wPassword);

			CConverter wEmail(q.fieldValue(3));
			SetEmail(wEmail);

			CConverter wFullName(q.fieldValue(4));
			SetFullName(wFullName);

			q.nextRow();

			return 0;
		}
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::FindUser(const char* pUserName, const char* pPassword)
{
	if(!pUserName || strlen(pUserName) > 100)
		return -1;

	if(!pPassword || strlen(pPassword) > 64)
		return -1;

	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
			return -1;

		CppSQLite3Statement stmt = db.compileStatement(_T("select nIndex from users where sUsername=? and sPassword=? and nDeleted != 1;"));

		stmt.bind(1, pUserName);
		stmt.bind(2, pPassword);

		CppSQLite3Query q = stmt.execQuery();

		if(!q.eof())
		{
			return (q.getIntField(0));
		}
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::FindUsername(const char* pUserName)
{
	if(!pUserName || strlen(pUserName) > 100)
		return -1;

	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
			return -1;


		CppSQLite3Statement stmt = db.compileStatement(_T("select nIndex from users where sUsername=?;"));

		stmt.bind(1, pUserName);

		CppSQLite3Query q = stmt.execQuery();

		if(!q.eof())
		{
			return (q.getIntField(0));
		}
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::Insert()
{
	if(!m_sUserName || strlen(m_sUserName) > 100)
		return -1;

	if(!m_sPassword || strlen(m_sPassword) > 64)
		return -1;

	if(!m_sEmail || strlen(m_sEmail) > 100)
		return -1;

	if(!m_sFullName || strlen(m_sFullName) > 100)
		return -1;
	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
		{
			db.execDML(_T("create table users(nIndex integer primary key, sUsername char(100), sPassword char(64), sEmail char(100), sFullname char(100), nDeleted int);"));
		}

		CppSQLite3Statement stmt = db.compileStatement(_T("insert into users (sUsername, sPassword, sEmail, sFullname, nDeleted)"
		" values (?, ?, ?, ?, ?);"));

		stmt.bind(1, m_sUserName);
		stmt.bind(2, m_sPassword);
		stmt.bind(3, m_sEmail);
		stmt.bind(4, m_sFullName);
		stmt.bind(5, m_nDeleted);


		int nRows = stmt.execDML();

		m_nID = (int)db.lastRowId();

		return nRows;
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::Update()
{
	if(m_nID <= 0)
		return -1;

	if(!m_sPassword || strlen(m_sPassword) > 64)
		return -1;

	if(!m_sEmail || strlen(m_sEmail) > 100)
		return -1;

	if(!m_sFullName || strlen(m_sFullName) > 100)
		return -1;

	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists("users"))
			return -1;

		CppSQLite3Statement stmt;

		if(strlen(m_sPassword) > 0)
		{
			stmt = db.compileStatement(_T("update users set sPassword=?, sEmail=?, sFullname=? where nIndex=?;"));
			stmt.bind(1, m_sPassword);
			stmt.bind(2, m_sEmail);
			stmt.bind(3, m_sFullName);
			stmt.bind(4, m_nID);
		}
		else
		{
			stmt = db.compileStatement(_T("update users set sEmail=?, sFullname=? where nIndex=?;"));
			stmt.bind(1, m_sEmail);
			stmt.bind(2, m_sFullName);
			stmt.bind(3, m_nID);
		}

		int nRows = stmt.execDML();
		return nRows;
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

void CDBUser::EncryptPassword(char* pKey, int nKeySize, const char* _pUsername, const char* _pPassword)
{
	if(!pKey)
		return;

	if(nKeySize == 0)
		return;

	if(!_pUsername || strlen(_pUsername) > 100)
		return;
	if(!_pPassword || strlen(_pPassword) > 64)
		return;

	CSHA1 sha1;

	CConverter aUsername(_pUsername);
	CConverter aPassword(_pPassword);

	sha1.Update((UINT_8*)(char*)(const char*)aUsername, strlen(aUsername));
	sha1.Update((UINT_8*)(char*)(const char*)aPassword, strlen(aPassword));
	sha1.Final();

	char* aKey = new char[nKeySize];
	memset(aKey, 0, nKeySize * sizeof(char));
	sha1.ReportHash(aKey, CSHA1::REPORT_HEX);

	CConverter _pKey(aKey);
	strncpy(pKey, _pKey, nKeySize);

	delete aKey;
}

int CDBUser::Delete(int nUserID)
{
	if(nUserID <= 0)
		return -1;

	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
			return false;

		CppSQLite3Statement stmt = db.compileStatement(_T("update users set nDeleted=1 where nIndex=?;"));
		stmt.bind(1, nUserID);
		stmt.execDML();

		return 0;
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::GetUserCount()
{
	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
			return 0;

		int nRet = db.execScalar(_T("select count(*) from users where nDeleted = 0"));

		return nRet;
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::LoadUsers(CDBUser**& pUsers)
{
	int nUserCount = GetUserCount();
	if(nUserCount <= 0)
		return 0;

	pUsers = new CDBUser * [nUserCount];

	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
			return 0;

		CppSQLite3Query q = db.execQuery(_T("select * from users where nDeleted = 0;"));

		int i = 0;

		while(!q.eof())
		{
			CDBUser* pUser = new CDBUser;

			pUser->SetID(q.getIntField(0));

			CConverter wUsername(q.fieldValue(1));
			pUser->SetUsername(wUsername);

			CConverter wPassword(q.fieldValue(2));
			pUser->SetPassword(wPassword);

			CConverter wEmail(q.fieldValue(3));
			pUser->SetEmail(wEmail);

			CConverter wFullName(q.fieldValue(4));
			pUser->SetFullName(wFullName);

			pUsers[i++] = pUser;

			q.nextRow();
		}

		return nUserCount;
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::ResetPassword(int nUserID, const char* sPwd)
{
	if(nUserID <= 0)
		return -1;

	if(!sPwd || strlen(sPwd) > 64)
		return -1;

	try
	{
		CppSQLite3DB db;
		db.Open(m_sDB);

		if(!db.tableExists(_T("users")))
			return 0;

		CppSQLite3Statement stmt = db.compileStatement(_T("update users set sPassword=? where nIndex=? and nDeleted='0';"));

		stmt.bind(1, sPwd);
		stmt.bind(2, nUserID);

		stmt.execDML();

		return 0;
	}
	catch(CppSQLite3Exception& e)
	{
		CConverter wMsg(e.errorMessage());
		CDBUser::Log(wMsg);
	}

	return -1;
}

int CDBUser::GetID()
{
	return m_nID;
}

char* CDBUser::GetUsername()
{
	return m_sUserName;
}

char* CDBUser::GetPassword()
{
	return m_sPassword;
}

char* CDBUser::GetEmail()
{
	return m_sEmail;
}

char* CDBUser::GetFullName()
{
	return m_sFullName;
}

bool CDBUser::IsDeleted()
{
	return m_nDeleted;
}

void CDBUser::SetID(int _nID)
{
	if(_nID <= 0)
		return;

	m_nID = _nID;
}

void CDBUser::SetUsername(const char* _sUserName)
{
	if(!_sUserName || strlen(_sUserName) > 100)
		return;

	delete m_sUserName;
	m_sUserName = strdup(_sUserName);
}

void CDBUser::SetPassword(const char* _sPassword)
{
	if(!_sPassword || strlen(_sPassword) > 64)
		return;

	m_sPassword = strdup(_sPassword);
}

void CDBUser::SetEmail(const char* _sEmail)
{
	if(!_sEmail || strlen(_sEmail) > 100)
		return;

	delete m_sEmail;
	m_sEmail = strdup(_sEmail);
}

void CDBUser::SetFullName(const char* _sFullName)
{
	if(!_sFullName || strlen(_sFullName) > 100)
		return;

	delete m_sFullName;
	m_sFullName = strdup(_sFullName);
}

void CDBUser::SetDeleted(bool b)
{
	m_nDeleted = b;
}

void CDBUser::Log(const char* msg)
{
	if(m_pLogFunc)
		m_pLogFunc(msg);
}

void CDBUser::SetLogFunc(LOG_ROUTINE _pLogFunc)
{
	m_pLogFunc = _pLogFunc;
}

void CDBUser::SetLastAccess(int _nLastAccess)
{
	m_nLastAccess = _nLastAccess;
}

void CDBUser::SetWritten(int _nWritten)
{
	m_nWritten = _nWritten;
}

void CDBUser::SetRead(int _nRead)
{
	m_nRead = _nRead;
}

void CDBUser::SetWriteTime(int _nWriteTime)
{
	m_nWriteTime = _nWriteTime;
}

void CDBUser::SetReadTime(int _nReadTime)
{
	m_nReadTime = _nReadTime;
}

int CDBUser::GetLastAccess()
{
	return m_nLastAccess;
}

int CDBUser::GetWritten()
{
	return m_nWritten;
}

int CDBUser::GetRead()
{
	return m_nRead;
}

int CDBUser::GetWriteTime()
{
	return m_nWriteTime;
}

int CDBUser::GetReadTime()
{
	return m_nReadTime;
}
