////////////////////////////////////////////////////////////////////////////////
//	Copyright (c) CitizenLab, 2006. All Rights Reserved.
//	The latest version of this code is available at http://psiphon.civisec.org/
//
//	This software is open source; you can redistribute it and/or modify it 
//	under the terms of the GNU General Public License as published by the 
//	Free Software Foundation; either version 2 of the License, or (at your 
//	option) any later version.
//
//	This program is distributed WITHOUT ANY WARRANTY; without even the 
//	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//	See the GNU General Public License for more details at:
//	http://psiphon.civisec.org/
//	
//	This General Public License does NOT permit incorporating this software 
//	into proprietary programs.
////////////////////////////////////////////////////////////////////////////////

#include "parsedUrl.h"
#include "strfunc.h"
#include "appWeb/appWeb.h"
#include <string.h>

CParsedUrl::CParsedUrl()
{
	Init();
}

CParsedUrl::CParsedUrl(char* _sUrl)
{
	Init();
	Parse(_sUrl);
}

CParsedUrl::~CParsedUrl()
{
	delete	m_sUrl;
	delete	m_sProto;
	delete	m_sHost;
	delete	m_sPath;
	delete	m_sQuery;
}

void CParsedUrl::Init()
{
	m_sUrl = strdup("");
	m_sProto = strdup("");
	m_sHost = strdup("");
	m_sPath = strdup("");
	m_sQuery = strdup("");
	m_nPort = 0;
	m_bSSL = false;
	m_nParseError = 0;
}

int CParsedUrl::Parse(const char* _sUrl)
{
	SetURL(_sUrl);

	const char* curr_p = m_sUrl;

	curr_p = ExtractProtocol(curr_p);
	if(curr_p == NULL)
		return -1;

//extract host
	curr_p = SkipToHost(curr_p); //E.F. FUTURE(parse username and password)

	const char* host_b = curr_p;
	const char* host_e = strpbrk(curr_p, ":/?#");

	if(!host_e)
		host_e = strchr(curr_p, '\0');


	if(host_e == host_b)
	{
		SetError(BAD_HOST);	
		return -1;
	}

	SetHost(host_b, host_e - host_b);

	curr_p = host_e;
//

//extract port
	if(*curr_p == ':') //port extraction routine
	{
		const char* port_b = NULL;
		const char* port_e = NULL;
		const char* pp = NULL;

		++curr_p;
		port_b = curr_p;
		
		port_e = strpbrk(curr_p, "/?#");

		if(!port_e)
			port_e = strchr(curr_p, '\0');

		if(port_b != port_e)
		{
			for(m_nPort = 0, pp = port_b;pp < port_e;pp++)
			{
				if(!isdigit(*pp))
				{
					SetError(BAD_PORT_NUMBER);
					return -1;
				}
				m_nPort = 10 * m_nPort + (*pp - '0');

				if(m_nPort > 65535)
				{
					SetError(BAD_PORT_NUMBER);
					return -1;
				}
			}
		}

		curr_p = port_e;
	} 

	//extract path
	if(*curr_p == '/')
	{
		const char* path_b = curr_p;
		const char* path_e = strpbrk(curr_p, "?#");

		if(!path_e)
			path_e = strchr(curr_p, '\0');

		SetPath(path_b, path_e - path_b);

		curr_p = path_e;
	}
	else
	{
		SetPath("/");
	}

	//extract query
	if(*curr_p == '?')
	{
		++curr_p;
		
		const char* query_b = curr_p;
		
		const char* query_e = strpbrk(curr_p, "#");

		if(!query_e)
			query_e = strchr(curr_p, '\0');

		SetQuery(query_b, query_e - query_b);

		curr_p = query_e;
	}
	//
	return 0;
}

const char* CParsedUrl::ExtractProtocol(const char* pURL)
{
	if(StartsWith(pURL, "http://"))
	{
		SetProtocol("http");
		SetPort(80);
		SetSSL(false);

		return (pURL + strlen(GetProtocol()) + strlen("://")); 
	}
	
	if(StartsWith(pURL, "https://"))
	{
		SetProtocol("https");
		SetPort(443);
		SetSSL(true);

		return (pURL + strlen(GetProtocol()) + strlen("://")); 
	}

	else //no recognizable protocol header
	{
		char* s = strstr(pURL, "://");  //ftp?
		if(s)
		{
			char* pProto = strndup(pURL, s - pURL);
			SetProtocol(pProto);	
			delete pProto;
			return s + strlen("://");
		}
		else
		{
			int nSize = strlen("http://") + strlen(pURL) + 1;
			char* urlBuf = new char[nSize];	

			int nLen = _snprintf(urlBuf, nSize, "http://%s", pURL);
			if(nLen <= 0 || nLen >= MPR_HTTP_MAX_URL)
			{
				delete urlBuf;
				return NULL;
			}
			
			SetURL(urlBuf);
			SetProtocol("http");
			SetPort(80);
			SetSSL(false);
		
			delete urlBuf;
			return GetURL() + strlen("http://");
		}
	}
}

const char* CParsedUrl::SkipToHost(const char* url)
{
	const char* p = (const char*)strpbrk(url, "@/?#");

	if(!p || *p != '@')
		return url;

	return p + 1;
}

void CParsedUrl::SetError(eParseError err)
{
	m_nParseError = err;
}

const char* CParsedUrl::GetURL() const
{
	return m_sUrl;
}   			  	

void CParsedUrl::SetURL(const char* pURL)
{
	delete m_sUrl;
	m_sUrl = strdup(pURL);
}   			  	

const char* CParsedUrl::GetProtocol() const
{
	return m_sProto;
}  

void CParsedUrl::SetProtocol(const char* sProto)
{
	delete m_sProto;
	m_sProto = strdup(sProto);
}

int CParsedUrl::GetPort() const
{
	return m_nPort;
}

void CParsedUrl::SetPort(int nPort)
{
	m_nPort = nPort;
}

const char* CParsedUrl::GetHost() const
{
	return m_sHost;
}  

void CParsedUrl::SetHost(const char* sHost, int nLength)
{
	delete m_sHost;
	m_sHost = strndup(sHost, nLength);
}

const char* CParsedUrl::GetPath() const
{
	return m_sPath;
}  

void CParsedUrl::SetPath(const char* sPath, int nLength)
{
	delete m_sPath;

	if(nLength == 0)
		m_sPath = strdup(sPath);
	else
		m_sPath = strndup(sPath, nLength);
}

const char* CParsedUrl::GetQuery() const
{
	return m_sQuery;
}  

void CParsedUrl::SetQuery(const char* sQuery, int nLength)
{
	delete m_sQuery;
	m_sQuery = strndup(sQuery, nLength);
}

bool CParsedUrl::GetSSL() const
{
	return m_bSSL;
} 

void CParsedUrl::SetSSL(bool bSSL)
{
	m_bSSL = bSSL;
}

int CParsedUrl::GetError() const
{
	return m_nParseError;
}   			 	
