////////////////////////////////////////////////////////////////////////////////
//	Copyright (c) CitizenLab, 2006. All Rights Reserved.
//	The latest version of this code is available at http://psiphon.civisec.org/
//
//	This software is open source; you can redistribute it and/or modify it 
//	under the terms of the GNU General Public License as published by the 
//	Free Software Foundation; either version 2 of the License, or (at your 
//	option) any later version.
//
//	This program is distributed WITHOUT ANY WARRANTY; without even the 
//	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//	See the GNU General Public License for more details at:
//	http://psiphon.civisec.org/
//	
//	This General Public License does NOT permit incorporating this software 
//	into proprietary programs.
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <vector>
#include <string>
#include	"mpr.h"
#include	"shared.h"

/////////////////////////////////// Defines ////////////////////////////////////
//
//	Constants
//
#define MPR_HTTP_CLIENT_NAME	"mbedthis-client/" BLD_VERSION

#if BLD_FEATURE_SQUEEZE
#define MPR_HTTP_CLIENT_BUFSIZE	(512)			// General I/O buffer size
#define MPR_HTTP_MAX_CONTENT	(32 * 1024)		// Max data on client request
#else
#define MPR_HTTP_CLIENT_BUFSIZE	(1024 * 4)		// General I/O buffer size
#define MPR_HTTP_MAX_CONTENT	(64 * 1024)		// Max data on client request
#endif

#define MPR_HTTP_CLIENT_RETRIES	(2)
#define MPR_HTTP_CLIENT_TIMEOUT	(45 * 1000)		// A bit shorter than server 

// 
//	Input states
// 
#define MPR_HTTP_CLIENT_START	0x1			// Starting state 
#define MPR_HTTP_CLIENT_HEADER	0x2			// Ready to read first line 
#define MPR_HTTP_CLIENT_CHUNK	0x4			// Reading chunk length
#define MPR_HTTP_CLIENT_CONTENT	0x8			// Reading posted content
#define MPR_HTTP_CLIENT_DONE	0x20		// Processing complete

// 
//	Client flags
// 
#define MPR_HTTP_KEEP_ALIVE		0x1			// Keep connection alive for more 
#define MPR_HTTP_TERMINATED		0x2			// Request prematurely terminated
#define MPR_HTTP_COMPLETE		0x4			// Request complete
#define MPR_HTTP_INPUT_CHUNKED	0x8			// Using HTTP/1.1 chunked
#define MPR_HTTP_END_CHUNK_DATA	0x10		// End of chunked data

////////////////////////////////// MaClientEx ////////////////////////////////////

class MaClientEx;

typedef void		(*MaClientExProc)(MaClientEx* cp, void* arg);

class CLocker
{
	MprMutex* 	m_mutex;		

public:
	CLocker(MprMutex* _mutex) : m_mutex(_mutex)
	{
		if(m_mutex)
			m_mutex->lock();
	}

	~CLocker()
	{
		if(m_mutex)
			m_mutex->unlock();
	}
};

class CSocketLocker
{
	MprSocket* 	m_sock;		

public:
	CSocketLocker(MprSocket* _sock) : m_sock(_sock)
	{
		if(m_sock)
			m_sock->lock();
	}

	~CSocketLocker()
	{
		if(m_sock)
			m_sock->unlock();
	}
};

class MaClientEx : public MprLink
{
	MprStr						m_serverAlgorithm;
	MprStr						m_serverDomain;
	MprStr						m_serverNonce;
	MprStr						m_serverOpaque;
	MprStr						m_serverRealm;
	MprStr						m_serverQop;
	MprStr						m_serverStale;
	MprStr						m_serverAuthType;		// Basic or Digest

	int							m_authNc;
	MprStr						m_authCnonce;

	int							m_contentLength;		// Length of content data 
	int							m_contentRemaining;	// Remaining content data to read

	MprStr						m_currentHost;		// Kept-alive host
	int							m_currentPort;		// Kept-alive port

	MprStr						m_defaultHost;		// Default target host (if unspecified)
	int							m_defaultPort;		// Default target port (if unspecified)

	MprStr						m_errorMsg;			// Error message if failure 

	int							m_fd;					// Underlying socket file handle
	int							m_flags;				// Control flags 

	MprHashTable* 				m_headerValues;		// Headers keyword values
	MprStr						m_method;				// Request method "GET", "HEAD", "POST"

	MprBuf* 					m_inBuf;				// Response input data
	MprBuf* 					m_outBuf;				// Request output buffer

	MprStr						m_password;			// As the name says

	MprStr						m_proxyHost;			// Proxy host to connect via
	int							m_proxyPort;			// Proxy port to connect via

	MprStr						m_realm;				// Authorization realm

	int							m_responseCode;		// Url response code 
	MprStr						m_responseProto;		// Response protocol
	MprBuf* 					m_responseContent;	// Response data 
	MprBuf* 					m_responseHeader;		// Response header
	MprStr						m_responseText;		// Request response message 

	MaUrl						m_url;				// Request url

	MprSocket* 					m_sock;				// Underlying socket handle
	int							m_state;				// Read handler state 

	int							m_userFlags;			// User flags (same set as flags)
	MprStr						m_user;				// User account name
	MprStr						m_userHeaders;		// User headers

	MprCond* 					m_GotStuffCond;		// Signalled when request is complete

	MprMutex* 					m_mutex;				// Mutli-thread sync

	std::vector <std::string>	m_allowedVector;

public:
								MaClientEx();
								~MaClientEx();

	int							sendRequest(char* host, int port, MprBuf* hdrBuf, char* post, int len);

	int getFd()
	{
		return m_fd;
	};
	char*	getHeaderVar(char* key);
	char*	getHost();
	void	getParsedUrl(MaUrl** url);
	int		getPort();
	int		getResponseCode();
	char*	getResponseContent(int* contentLen);
	char*	getResponseHeader();

	MprSocket* getSock()
	{
		return m_sock;
	};

	void	resetAuth();
	void	setAuth(char* realm, char* user, char* password);
	void	setHost(char* host);
	void	setKeepAlive(bool on);
	void	setPort(int num);
	void	setProxy(char* host, int port);
	void	setRetries(int count);
	void	setTimeout(int timeout);
	void	setUserHeaders(char* headers);
	int		traceRequest(char* url);
	void	allowContentType(std::string allowed_content);	
	bool	IsAllowedContentType(char* cont_type);

	void lock()
	{
		m_mutex->lock();
	}
	void unlock()
	{
		m_mutex->unlock();
	}

public:
	int		readEvent();
	int		getState();
	void	signalGotStuff();

private:
	int		getFlags();
	int		inUse(int adjustment);
	bool	isDisposed();

private:
	int		parseAuthenticate(char* authDetails);
	int		parseFirst(char* line);
	int		parseHeader(char* line);
	int		processResponseData();
	void	finishRequest(bool closeSocket);
	void	reset();

	int		SocketGetFd();
	bool	SocketGetEOF();
	void	SocketReset();
	int		SocketInit(char* host, int port);
	int		SocketSetBufSize(int sendSize, int recvSize);
	bool	SocketCreate();
	int		SocketOpenClient(char* hostName, int portNum, int flags);
	void	SocketShutdown();
	void	SocketSetCallBack(MprSocketIoProc fn, void* data, void* data2, int mask, int pri = MPR_NORMAL_PRIORITY);
	int		SocketWrite(char* buf, int len);
	int		SocketRead(char* buf, int len);
};
