////////////////////////////////////////////////////////////////////////////////
//	Copyright (c) CitizenLab, 2006. All Rights Reserved.
//	The latest version of this code is available at http://psiphon.civisec.org/
//
//	This software is open source; you can redistribute it and/or modify it 
//	under the terms of the GNU General Public License as published by the 
//	Free Software Foundation; either version 2 of the License, or (at your 
//	option) any later version.
//
//	This program is distributed WITHOUT ANY WARRANTY; without even the 
//	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//	See the GNU General Public License for more details at:
//	http://psiphon.civisec.org/
//	
//	This General Public License does NOT permit incorporating this software 
//	into proprietary programs.
////////////////////////////////////////////////////////////////////////////////

#include "appWeb/appWeb.h"
#include "HTMLFlexLexer.h"
#include "CSSFlexLexer.h"
#include "tokensDefs.h"

#include "psiphonContFilter.h"
#include <strstream>
#include "psiphonHandler.h"
#include "SetupInfo.h"
#include "strfunc.h"
#include "psiphonUrl.h"

const tagData CPsiphonContFilter::m_tags[] =
{
	tagData("body", "background", IMGSRC), tagData("table", "background", IMGSRC), tagData("td", "background", IMGSRC), tagData("tr", "background", IMGSRC), tagData("a", "href", NOFRAME), tagData("link", "href", NOFRAME), tagData("base", "href", NOFRAME), tagData("area", "href", NOFRAME), tagData("img", "src", IMGSRC), tagData("img", "lowsrc", IMGSRC), tagData("input", "src", IMGSRC), tagData("script", "src", NOFRAME), tagData("iframe", "src", FRAME), tagData("frame", "src", FRAME), tagData("embed", "src", NOFRAME), tagData("bgsound", "src", NOFRAME), tagData("form", "action", NOFRAME)
};
const char* CPsiphonContFilter::jsEvents[] =
{
	"onabort", "onblur", "onchange", "onclick", "ondblclick", "onerror", "onfocus", "onkeydown", "onkeypress", "onkeyup", "onload", "onmousedown", "onmousemove", "onmouseout", "onmouseover", "onmouseup", "onreset", "onresize", "onselect", "onsubmit", "onunload", "onreadystatechange"
};

CPsiphonContFilter::CPsiphonContFilter(PsiphonHandler* _pSh, bool bStripJS, bool bStripObjects, bool bStripImages)
{
	m_pSh = _pSh;
	m_bStripJS = bStripJS;
	m_bStripObjects = bStripObjects;
	m_bStripImages = bStripImages;
}

CPsiphonContFilter::~CPsiphonContFilter()
{
}

int CPsiphonContFilter::IsJS()
{
	char* sTemp = mprStrLower(strdup(m_cur_content));
	if(StartsWith(sTemp, "javascript:"))
	{
		delete sTemp;	
		return 1;
	}
	delete sTemp;	

	for(int i = 0;i < sizeof(jsEvents) / sizeof(jsEvents[0]);i++)
	{
		if(!stricmp(m_cur_tag_attr, jsEvents[i]))
			return 1;
	}
	return 0;
}

int CPsiphonContFilter::IfRewrite()
{
	if(m_n_cur_token == T_META_REFRESH_URL) //check if meta-refresh URL
		return 1;

	for(int i = 0;i < sizeof(m_tags) / sizeof(m_tags[0]);i++)
	{
		if(!stricmp(m_cur_tag_name, m_tags[i].name) && !stricmp(m_cur_tag_attr, m_tags[i].attr))
		{
			return m_tags[i].rewrite_val;
		}
	}

	return 0;
}

int CPsiphonContFilter::PutCSSContent(char* baseurl)
{
	CPsiphonURL URL(m_cur_content, baseurl);

	char* transformed_url = NULL;
	int n_chars = 0;

	if(URL.TransformURL(transformed_url))
	{
		const char* sFetchURL = URL_FETCH;
		int nSize = strlen(m_pSh->m_pNodeName) + 1 + strlen(sFetchURL) + strlen(transformed_url) + 1;

		char* buf = new char[nSize];
		int nLen = _snprintf(buf, nSize, "/%s%s%s", m_pSh->m_pNodeName, sFetchURL, transformed_url);
		if(nLen <= 0 || nLen >= nSize)
		{
			delete buf;
			return -1;
		}

		n_chars = m_pSh->Write(buf);

		delete buf;

		free(transformed_url);

		return n_chars;
	}

	return m_pSh->Write(m_cur_content);
}

int CPsiphonContFilter::PutHTMLContent(char* baseurl)
{
	const char* sFetchURL = NULL;

	int nRewriteVal = IfRewrite();

	if(nRewriteVal)
	{
		if(nRewriteVal == FRAME)
			sFetchURL = URL_FETCHI;
		else
			sFetchURL = URL_FETCH;

		if((nRewriteVal == IMGSRC) && m_bStripImages)
			return m_pSh->Write("javascript:void(0);");

		CPsiphonURL URL(m_cur_content, baseurl);

		char* transformed_url = NULL;
		int n_chars = 0;

		if(URL.TransformURL(transformed_url))
		{
			int nSize = 1 + strlen(m_pSh->m_pNodeName) + strlen(sFetchURL) + strlen(transformed_url) + 1;

			char* buf = new char[nSize];
			int nLen = _snprintf(buf, nSize, "/%s%s%s", m_pSh->m_pNodeName, sFetchURL, transformed_url);
			if(nLen <= 0 || nLen >= nSize)
			{
				delete buf;
				return -1;
			}

			n_chars = m_pSh->Write(buf);

			delete buf;
		}
		else
		{
			n_chars = m_pSh->Write(m_cur_content);
		}

		free(transformed_url);

		return n_chars;
	}

	return m_pSh->Write(m_cur_content);
}

int CPsiphonContFilter::ProcessContent(std::istrstream* sStream, char* baseurl)
{
	m_n_cur_token = 0;
	char* base_href_url = NULL;
	int n_prev_token = 0;
	*m_cur_tag_attr = 0;
	*m_cur_tag_name = 0;

	FlexLexer* HTMLlexer = new HTMLFlexLexer();

	HTMLlexer->yyrestart(sStream);

	int nLen = 0;

	while((m_n_cur_token = HTMLlexer->yylex()) != 0)
	{
		m_cur_content = (char*)HTMLlexer->YYText();

		switch(m_n_cur_token)
		{
		case T_BEGIN_TAG:
			m_pSh->Write(m_cur_content);
			break;
		case T_END_TAG:
			m_pSh->Write(m_cur_content);
			if(stricmp(m_cur_tag_name, "base") == 0)
			{
				m_pSh->Write("-->");
			}
			*m_cur_tag_name = 0;
			*m_cur_tag_attr = 0;
			break;
		case T_CLOSE_TAG:
			m_pSh->Write(m_cur_content);
			if(stricmp(m_cur_tag_name, "base") == 0)
			{
				m_pSh->Write("-->");
			}
			*m_cur_tag_name = 0;
			*m_cur_tag_attr = 0;
			break;
		case T_TAG_NAME:
			nLen = _snprintf(m_cur_tag_name, MAX_TAG_NAME_SIZE, "%s", m_cur_content);
			if(nLen <= 0 || nLen >= MAX_TAG_NAME_SIZE)
			{
				delete HTMLlexer;			
				return -1;
			}

			if(stricmp(m_cur_tag_name, "frameset") == 0 && m_pSh->m_bTopBar)
			{
				m_pSh->Write("frameset rows=\"30px, *\"><frame src = \"/topbar-frame/\" marginwidth=0 marginheight=0><");
				m_pSh->m_bTopBar = false;
			}
			else if(stricmp(m_cur_tag_name, "base") == 0)
			{
				m_pSh->Write("!--");
			}
			m_pSh->Write(m_cur_content);

			break;
		case T_TAG_PARAMETER:
			m_pSh->Write(m_cur_content);

			nLen = _snprintf(m_cur_tag_attr, MAX_TAG_ATTR_SIZE, "%s", m_cur_content);
			if(nLen <= 0 || nLen >= MAX_TAG_ATTR_SIZE)
			{
				delete HTMLlexer;			
				return -1;
			}

			break;
		case T_TAG_PARAMETER_VALUE:
			if(stricmp(m_cur_tag_attr, "style") == 0)
			{
				char* decoded = DecodeAttrVal(m_cur_content);
				std::istrstream* sStyleString = new std::istrstream(decoded);
				ProcessCSS(sStyleString, baseurl);
				delete sStyleString;
			}
			else if(!stricmp(m_cur_tag_name, "base") && !stricmp(m_cur_tag_attr, "href"))
			{
				CPsiphonURL* pUrl = new CPsiphonURL(m_cur_content, baseurl);
				pUrl->TransformURL(base_href_url);
				baseurl = base_href_url;
				delete pUrl;
			}
			else if(IsJS() && m_bStripJS)
			{
				m_pSh->Write("javascript:void(0);");
			}
			else
				PutHTMLContent(baseurl);
			break;
		case T_META_REFRESH_URL:
			PutHTMLContent(baseurl);
			break;
		case T_SCRIPT:
			*m_cur_tag_name = 0;
			*m_cur_tag_attr = 0;
			if(n_prev_token != m_n_cur_token && m_bStripJS)
				m_pSh->Write("script>/*removed by psiphon*/</script>");
			else if(!m_bStripJS)
				m_pSh->Write(m_cur_content);
			break;
		case T_NOSCRIPT:
			if(!m_bStripJS)
				m_pSh->Write(m_cur_content);
			break;
		case T_NOSCRIPT_END:
			if(!m_bStripJS)
				m_pSh->Write(m_cur_content);
			break;
		case T_IN_STYLE:
			m_pSh->Write(m_cur_content);
			ProcessCSS(sStream, baseurl);
			HTMLlexer->yyrestart(sStream);
			break;
		case T_OBJECT:
			if(!m_bStripObjects)
				m_pSh->Write(m_cur_content);
			break;
		default:
			m_pSh->Write(m_cur_content);
			break;
		}
		n_prev_token = m_n_cur_token;
	}

	delete base_href_url;
	delete HTMLlexer;

	return 0;
}

int CPsiphonContFilter::ProcessCSS(std::istrstream* sStream, char* baseurl)
{
	int lex_state = 0;

	FlexLexer* CSSlexer = new CSSFlexLexer();

	CSSlexer->yyrestart(sStream);

	while((lex_state = CSSlexer->yylex()) != 0)
	{
		m_cur_content = (char*)CSSlexer->YYText();

		if(lex_state == T_STYLE_URL)
			PutCSSContent(baseurl);
		else
			m_pSh->Write(m_cur_content);

		if(lex_state == T_STYLE_END)
			break;
	}

	delete CSSlexer;
	return 0;
}
void CPsiphonContFilter::DecodeHTMLEntity(char** val, std::strstream* stream)
{
	long n = 0;
	if(**val == 'x' || **val == 'X') //hexademical
	{
		(*val)++;
		while(isxdigit(**val))
		{
			if(**val <= '9')
				n = 16 * n + **val - '0';
			else if(**val <= 'F')
				n = 16 * n + **val - 'A' + 10;
			else
				n = 16 * n + **val - 'a' + 10;
			if(n < 0)
				n = -1;
			(*val)++;
		}
	}
	else //decimal
	{
		n = **val - '0';
		(*val)++;
		while(isdigit(**val))
		{
			n = 10 * n + **val - '0';
			if(n < 0)
				n = -1;
			(*val)++;
		}
	}
	if(**val == ';')
		(*val)++;
	if(n >= 0x20 && n <= 0x7F && n!=0x23 && n!=0x26)
		*stream << (char)n;
	else if(n >= 0)
		*stream << "&#x" << std::setbase(16) << n << ";";
	else
		*stream << "&#x" << std::setbase(16) << (int)'?' << ";";  //unknown
} 
char* CPsiphonContFilter::DecodeAttrVal(char* content)
{
	char* found = NULL;
	std::strstream out;

	found = content;
	while(*found)
	{
		if(*found == '&')
		{
			found++;
			if(*found != '#')
			{
				out << *found; 
				continue;
			}
			found++;
			DecodeHTMLEntity(&found, &out);
		}
		else
		{
			out << *found;
			found++;
		}
	}    
	out << '\0';
	return out.str();
}
