////////////////////////////////////////////////////////////////////////////////
//	Copyright (c) CitizenLab, 2006. All Rights Reserved.
//	The latest version of this code is available at http://psiphon.civisec.org/
//
//	This software is open source; you can redistribute it and/or modify it 
//	under the terms of the GNU General Public License as published by the 
//	Free Software Foundation; either version 2 of the License, or (at your 
//	option) any later version.
//
//	This program is distributed WITHOUT ANY WARRANTY; without even the 
//	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//	See the GNU General Public License for more details at:
//	http://psiphon.civisec.org/
//	
//	This General Public License does NOT permit incorporating this software 
//	into proprietary programs.
////////////////////////////////////////////////////////////////////////////////

#define		IN_SIMPLE_HANDLER 1

#include	"appWeb/appWeb.h"
#include	"psiphonHandler.h"
#include	"psiphonContFilter.h"
#include	"psiphonUrl.h"
#include	"psiphonUtils.h"
#include	"User.h"
#include	"SHA1.h"
#include	"logDetails.h"
#include	"timefunc.h"
#include	"bookmarks.h"
#include	"cookies.h"
#include	"sessions.h"
#include	"SetupInfo.h"
#include 	"strfunc.h"


////////////////////////////////////////////////////////////////////////
////////////////////////// PsiphonHandlerModule /////////////////////////
////////////////////////////////////////////////////////////////////////

//
//	Module entry point. This is only called when the module is loaded 
//	as a DLL.
//

int mprPsiphonHandlerInit(void* handle)
{
	if(maGetHttp() == 0)
	{
		return MPR_ERR_NOT_INITIALIZED;
	}

	new PsiphonHandlerModule(handle);
	return 0;
}

////////////////////////////////////////////////////////////////////////

///
///	Constructor. Called either above when the DLL is loaded or from the 
///	application main if loading statically.
///

PsiphonHandlerModule::PsiphonHandlerModule(void* handle) : MaModule("psiphonHandler", handle)
{
	simpleService = new PsiphonHandlerService();
}

////////////////////////////////////////////////////////////////////////

///
///	Module destructor
///

PsiphonHandlerModule::~PsiphonHandlerModule()
{
	delete simpleService;
}

////////////////////////////////////////////////////////////////////////
///////////////////////// PsiphonHandlerService /////////////////////////
////////////////////////////////////////////////////////////////////////

///
///	Constructor for the PsiphonHandler service. An instance is created 
///	for each host (default and virtual).
///

PsiphonHandlerService::PsiphonHandlerService() : MaHandlerService("psiphonHandler")
{
}

////////////////////////////////////////////////////////////////////////

///
///	Destructor for the PsiphonHandler. 
///

PsiphonHandlerService::~PsiphonHandlerService()
{
}

////////////////////////////////////////////////////////////////////////

///
///	Setup the PsiphonHandler service. One-time setup for a given host.
///

int PsiphonHandlerService::setup()
{
	return 0;
}

////////////////////////////////////////////////////////////////////////

///
///	New Handler factory. Create a new PsiphonHandler for an incoming 
///	HTTP request for a given server
///

MaHandler* PsiphonHandlerService::newHandler(MaServer*, MaHost*, char* extensions)
{
	return new PsiphonHandler(extensions);
}

////////////////////////////////////////////////////////////////////////

PsiphonHandler::PsiphonHandler(char* extensions) : MaHandler("psiphonHandler", extensions, MPR_HANDLER_GET | MPR_HANDLER_HEAD | MPR_HANDLER_POST | MPR_HANDLER_MAP_VIRTUAL | MPR_HANDLER_EXTRA_PATH | MPR_HANDLER_NEED_ENV | MPR_HANDLER_TERMINAL | MPR_HANDLER_ALWAYS)
{
	Init();
}

PsiphonHandler::~PsiphonHandler()
{
	delete m_pPostBuf;
	free(m_pFetchURL);
	free(m_pNodeName);
	free(m_pRequestURL);

	Init();
}

void PsiphonHandler::Init()
{
	m_currentUser.Init();

	m_pPostBuf = NULL;
	m_pRq = NULL;
	m_bTopBar = true;

	m_pFetchURL = NULL;
	m_pNodeName = NULL;
	m_pRequestURL = NULL;
}

////////////////////////////////////////////////////////////////////////

///
///	For maximum speed in servicing requests, we clone a pre-existing 
///	handler when an incoming request arrives.
///

MaHandler* PsiphonHandler::cloneHandler()
{
	PsiphonHandler* ep = new PsiphonHandler(extensions);
	return ep;
}

////////////////////////////////////////////////////////////////////////

///
///	Called to setup any data structures before running the request
///

int PsiphonHandler::setup(MaRequest* rq)
{
	m_pRq = rq;

	MaLimits* limits = m_pRq->host->getLimits();
	mprAssert(m_pPostBuf == 0);

	m_pPostBuf = new MprBuf(MPR_HTTP_IN_BUFSIZE, limits->maxBody);

	CSetupInfo info;
	info.Load();

	m_bServerShowImages = info.GetShowImages();

	free(m_pNodeName);
	m_pNodeName = strdup(info.GetName());

	free(m_pRequestURL);
	m_pRequestURL = strdup(m_pRq->getOriginalUri());
	
	return 0;
}

////////////////////////////////////////////////////////////////////////

///
///	Called to see if this handler should process this request.
///	Return TRUE if you want this handler to match this request
///

int PsiphonHandler::matchRequest(MaRequest*, char*, int)
{
	return 1; //match all
}

////////////////////////////////////////////////////////////////////////

///
///	Receive any post data. Will be called after the run() method has 
///	been called and may be called multiple times. End of data is when 
///	remainingContent() == 0.
///

void PsiphonHandler::postData(MaRequest*, char* buf, int len)
{
	if(len < 0 && m_pRq->getRemainingContent() > 0)
	{
		m_pRq->finishRequest(MPR_HTTP_CLOSE);
		return;
	}

	int rc = m_pPostBuf->put((uchar*)buf, len);
	m_pPostBuf->addNull();

	if(rc != len)
	{
		m_pRq->finishRequest(MPR_HTTP_CLOSE);
		return;
	} 

	if(m_pRq->getRemainingContent() <= 0)
	{
		m_pRq->createQueryVars(m_pPostBuf->getStart(), m_pPostBuf->getLength());
		run(m_pRq);
	}
}

////////////////////////////////////////////////////////////////////////

///
///	Run the handler to service the request
///

int PsiphonHandler::run(MaRequest*)
{
	int flags = m_pRq->getFlags(); 

	if(flags & MPR_HTTP_POST_REQUEST && m_pRq->getRemainingContent() > 0)
	{
		//
		//	When all the post data is received the run method will be recalled
		//	by the postData method.
		//
		return MPR_HTTP_HANDLER_FINISHED_PROCESSING;
	}

	ServeRequest();

	unsigned long tStart = GetTime();

	m_pRq->flushOutput(MPR_HTTP_BACKGROUND_FLUSH, 0);

	unsigned long tEnd = GetTime();
	unsigned long tWriteTime = tEnd - tStart;
	m_currentUser.SetWriteTime(m_currentUser.GetWriteTime() + tWriteTime);

	LogRequest();

	m_pRq->finishRequest(1);

	return MPR_HTTP_HANDLER_FINISHED_PROCESSING;
}

//	examine method - GET or POST, otherwise serve "Bad Request"
int PsiphonHandler::ValidateRequest(char*& http_method)
{
	int nFlags = m_pRq->getFlags();

	if(nFlags & MPR_HTTP_GET_REQUEST)
	{
		http_method = "GET";
	}
	else if(nFlags & MPR_HTTP_POST_REQUEST)
	{
		http_method = "POST";
	}
	else
	{
		return RequestError(501);
	}

	return 0;
}

bool PsiphonHandler::IsIndexPage()
{
	const char* uri = m_pRq->getOriginalUri();

	if(!uri)
		return false;

	if(!strcmp(uri, "/index.html"))
		return true;

	if(!strcmp(uri, "/"))
		return true;

	while(strlen(uri) > 1 && StartsWith(uri, "//"))
		uri++;

	if(!strcmp(uri, "/"))
		return true;

	char* p_q = strchr(uri, '?');
	char* p_last_slash = strrchr(uri, '/');

	if(p_q != NULL)
	{
		if(p_q < p_last_slash)
			return false;
	}

	if(StartsWith(uri, "/?"))
		return true;

	if(StartsWith(uri, "/index.html?"))
		return true;

	return false;
}

void PsiphonHandler::ShowIndexPage()
{
	m_pRq->setResponseCode(200);
	m_pRq->setHeader("Content-type: text/html", false);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	CPsiphonUtils util(this);

	char* template_buffer = NULL;
	util.GetHTML("index.tpl.html", template_buffer);

	if(template_buffer)
	{
		Write(template_buffer);
		delete template_buffer;
	}
}

int PsiphonHandler::ValidateRequestURI()
{
	if(m_pRequestURL == NULL || *m_pRequestURL == 0)
		return NOTVALID;

	int nURLLen = strlen(m_pRequestURL);

	if(nURLLen <= 0 || nURLLen > MPR_HTTP_MAX_URL - 1)
		return NOTVALID;

	if(m_pNodeName == NULL || *m_pNodeName == 0)
		return NOTVALID;

	int nSize = strlen(m_pNodeName) + 3;

	char* sNodeURLPart = new char[nSize];
	int nLen = _snprintf(sNodeURLPart, nSize, "/%s/", m_pNodeName);
	if(nLen <= 0 || nLen >= nSize)
	{
		delete sNodeURLPart;
		return NOTVALID;
	}

	if(sNodeURLPart == NULL || *sNodeURLPart == 0)
	{
		return NOTVALID;
	}

	//redirect /<node>/ to /<node>/login/
	if(!strcmp(m_pRequestURL, sNodeURLPart))
	{
		delete sNodeURLPart;
		return LOGIN_REDIRECT;
	}

	//redirect /<node> to /<node>/login/
	nSize = strlen(m_pNodeName) + 2;

	char* sNodeURLPartNoSlash = new char[nSize];
	nLen = _snprintf(sNodeURLPartNoSlash, nSize, "/%s", m_pNodeName);
	if(nLen <= 0 || nLen >= nSize)
	{
		delete sNodeURLPart;
		delete sNodeURLPartNoSlash;
		return NOTVALID;
	}

	if(!strcmp(m_pRequestURL, sNodeURLPartNoSlash))
	{
		delete sNodeURLPart;
		delete sNodeURLPartNoSlash;
		return LOGIN_REDIRECT;
	}
	delete sNodeURLPartNoSlash;

	//check if index page in case if url doesn't start with /<node>/...
	if(StartsWith(m_pRequestURL, sNodeURLPart) == false)
	{
		delete sNodeURLPart;

		if(IsIndexPage() == true)
		{
			return INDEX;
		}
		else
			return NOTVALID;
	}
	delete sNodeURLPart;

	//examine /<url> part of original "/<node>/<url>" 
	char* request_uri = m_pRequestURL + strlen(m_pNodeName) + 1; 

	//serve files from disk
	if(StartsWith(request_uri, URL_SERVEFILE))
		return SERVEFILE;	

	//user login page
	if(StartsWith(request_uri, URL_LOGIN))
		return LOGIN;

	//redirect to login page
	if(!strcmp(request_uri, URL_LOGIN_NOSLASH))
		return LOGIN_REDIRECT;

	//user change password page
	if(!strcmp(request_uri, URL_CHANGEPWD))
		return CHANGEPWD;

	//redirect to change password
	if(!strcmp(request_uri, URL_CHANGEPWD_NOSLASH))
		return CHANGEPWD_REDIRECT;

	//send "AJAX" bookmarks
	if(!strcmp(request_uri, URL_GETBOOKMARKS))
		return GETBOOKMARKS;

	//add "AJAX" bookmarks
	if(StartsWith(request_uri, URL_ADDBOOKMARK))
		return ADDBOOKMARK;

	//delete "AJAX" bookmarks
	if(StartsWith(request_uri, URL_DELETEBOOKMARK))
		return DELETEBOOKMARK;

	//flag for main page request
	if(!strcmp(request_uri, URL_FETCH))
		return MAINPAGE;

	//flag for fetch request
	if(StartsWith(request_uri, URL_FETCH))
		return FETCH;

	//redirect to fetch
	if(!strcmp(request_uri, URL_FETCH_NOSLASH))
		return FETCH_REDIRECT;

	//flag for fetch request with no topbar
	if(StartsWith(request_uri, URL_FETCHI))
		return FETCHI;

	//redirect to fetch-i
	if(!strcmp(request_uri, URL_FETCHI_NOSLASH))
		return FETCHI_REDIRECT;

	//flag for framed page
	if(StartsWith(request_uri, URL_TOPBARFRAME))
		return TOPBARFRAME;

	//flag for setting language preferences
	if(StartsWith(request_uri, URL_SETLANG))
		return SETLANG;

	return NOTVALID;
}

int PsiphonHandler::ValidateRequestProtocol(const char* url)
{
	if(StartsWith(url, "http://"))
		return 0;

	return RequestError(501);
}

//redirect from http://site.com?q=foo to http://site.com/?q=foo
int PsiphonHandler::RedirectRequestURI(const MaUrl& maUrl, char*& request_uri)
{
	bool b_colon_present = false;
	int n_port_len = 0;

	if(request_uri[strlen(maUrl.proto) + strlen("://") + strlen(maUrl.host)] == ':')
		b_colon_present = true;

	if(b_colon_present)
	{
		unsigned int i = strlen(maUrl.proto) + strlen("://") + strlen(maUrl.host) + 1;
		while(i < (strlen(request_uri) + 1))
		{
			if(request_uri[i + n_port_len] == '?' || request_uri[i + n_port_len] == '/' || request_uri[i + n_port_len] == '#' || request_uri[i + n_port_len] == '\0')
				break;

			n_port_len++;
		}
		if(n_port_len > 0)
			n_port_len++;
	}

	if(request_uri[strlen(maUrl.proto) + strlen("://") + strlen(maUrl.host) + n_port_len] == '/')
		return 0;

	if(strlen(request_uri) >= MPR_HTTP_MAX_URL - strlen(m_pFetchURL) - strlen("://") - n_port_len)
	{
		return RequestError(400);
	}

	char* s_tail = request_uri + strlen(maUrl.proto) + strlen("://") + strlen(maUrl.host) + n_port_len;

	char urlBuf[MPR_HTTP_MAX_URL];	
	int nLen = 0;

	if(s_tail && *s_tail)
	{
		if(n_port_len)
			nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "%s%s://%s:%d/%s", m_pFetchURL, maUrl.proto, maUrl.host, maUrl.port, s_tail);
		else
			nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "%s%s://%s/%s", m_pFetchURL, maUrl.proto, maUrl.host, s_tail);
	}
	else
	{
		if(n_port_len)
			nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "%s%s://%s:%d/", m_pFetchURL, maUrl.proto, maUrl.host, maUrl.port);
		else
			nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "%s%s://%s/", m_pFetchURL, maUrl.proto, maUrl.host);
	}

	if(nLen <= 0 || nLen >= MPR_HTTP_MAX_URL)
	{
		return -1;
	}

	RedirectSpecial(urlBuf);

	return -1;
}

int PsiphonHandler::GetCookies(MprBuf& bufHeader, char* request_uri)
{
	//get session cookies from DB
	CDBUserCookies dbCookies;
	char* sCookies = dbCookies.BuildCookies(request_uri, m_currentUser.GetID());

	if(sCookies && *sCookies)
		bufHeader.putFmt("Cookie: %s", sCookies);

	delete sCookies;

	return 0;
}

int PsiphonHandler::PrepareHeader(MprBuf& bufHeader, char* http_method, char* request_uri, const MaUrl& maUrl)
{
	bufHeader.setBuf(MPR_HTTP_CLIENT_BUFSIZE, MPR_HTTP_MAX_BODY);

	if(strlen(maUrl.query))
		bufHeader.putFmt("%s %s?%s HTTP/1.0\r\n", http_method, maUrl.uri, maUrl.query);
	else
		bufHeader.putFmt("%s %s HTTP/1.0\r\n", http_method, maUrl.uri);

	bufHeader.putFmt("Host: %s\r\n", maUrl.host);
	bufHeader.putFmt("User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0)\r\n");
	bufHeader.putFmt("Referer: %s\r\n", request_uri);
	GetCookies(bufHeader, request_uri);

	bufHeader.putFmt("Connection: close\r\n");

	if(m_pPostBuf->getLength() > 0)
	{
		bufHeader.putFmt("Content-Length: %d\r\n", m_pPostBuf->getLength());
	}

	if(m_pPostBuf->getLength() > 0 && m_pPostBuf->getStart() && *m_pPostBuf->getStart() && m_pRq->getHeader()->contentMimeType)
	{
		bufHeader.putFmt("Content-Type: %s\r\n", m_pRq->getHeader()->contentMimeType);
	}

	bufHeader.put("\r\n");
	bufHeader.addNull();

	return 0;
}

int PsiphonHandler::StoreCookies(char* response_header, char* request_uri)
{
	char* startCookie = response_header; 
	char* endCookie = NULL; 
	std::vector <CCookie*> cookiesArray;
	CDBUserCookies dbCookies;

	while((startCookie = stristr(startCookie, "Set-Cookie: ")) != 0)
	{
		startCookie += strlen("Set-Cookie: ");
		endCookie = strchr(startCookie, '\n');

		if(endCookie > startCookie && endCookie - startCookie < 4096)
		{
			int nLength = endCookie - startCookie;

			char* cookieBuf = new char[nLength + 1];
			strncpy(cookieBuf, startCookie, nLength);
			cookieBuf[nLength] = '\0';

			CCookie* cookie = new CCookie();
			cookie->Setup(cookieBuf, request_uri);

			cookiesArray.push_back(cookie);
		}
	}

	if(cookiesArray.empty())
		return 0;

	dbCookies.StoreDBUserCookies(request_uri, cookiesArray, m_currentUser.GetID());

	std::vector<CCookie*>::iterator itPos = cookiesArray.begin();
	for(;itPos < cookiesArray.end();itPos++)
		delete * itPos;    

	cookiesArray.clear();

	return 0;
}

//	Examine the HTTP response HTTP code. 301-302 are redirects, 200 is success.
int PsiphonHandler::ExamineResponseCode(int nCode, MaClientEx* pClient, char* request_uri)
{
	if(nCode == 200)
		return 0;

	if(nCode <= 300 || nCode >= 400)
		return RequestError(nCode);

	//redirects
	char* redirect_location = pClient->getHeaderVar("location");

	if(redirect_location == NULL)
		return RequestError(400);

	//deal with incorrect redirects like "/page.html" that don't
	//have scheme / hostname in it.

	CPsiphonURL* pUrl = new CPsiphonURL(redirect_location, request_uri);

	char* redirect_path = NULL;
	pUrl->TransformURL(redirect_path);

	delete pUrl;

	//Getting content here only to invoke finishRequest which is 
	//private for MaClient. finishRequest closes socket on successful 
	//content retrieval 
	int contentLen = 0;
	pClient->getResponseContent(&contentLen);
	m_currentUser.SetRead(m_currentUser.GetRead() + contentLen);

	char urlBuf[MPR_HTTP_MAX_URL];	
	int nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "/%s%s%s", m_pNodeName, m_pFetchURL, redirect_path);
	if(nLen <= 0 || nLen >= MPR_HTTP_MAX_URL)
	{
		return RequestError(404);
	}

	free(redirect_path);
	redirect_path = urlBuf;

	m_pRq->setResponseCode(200);
	m_pRq->redirect(302, redirect_path);

	return -1;
}

int PsiphonHandler::RequestError(int nCode)
{
	m_pRq->setResponseCode(nCode);

	if(!IsLogged())
	{
		m_pRq->requestError(nCode, maGetHttpErrorMsg(nCode));
		return -1;
	}

	char sErrNo[64];
	int nLen = _snprintf(sErrNo, 64, "%d", nCode);

	const char* sFormat = "errors/%s.html";		
	int nErrFilenameLen = strlen(sFormat) + strlen(sErrNo) + 1;

	char* sErrFilename = new char[nErrFilenameLen];
	nLen = _snprintf(sErrFilename, nErrFilenameLen, sFormat, sErrNo);
	if(nLen <= 0 || nLen >= nErrFilenameLen)
	{
		delete sErrFilename;
		return -1;
	}

	char* sOut = NULL;
	CPsiphonUtils util(this);

	if(util.GetHTML(sErrFilename, sOut) != 0)
	{
		m_pRq->requestError(nCode, maGetHttpErrorMsg(nCode));
	}
	else
	{
		m_pRq->setHeader("Content-type: text/html", false);
		m_pRq->insertDataStream(m_pRq->getDynBuf());

		if(sOut)
			Write(sOut);

		if(m_bTopBar && m_pRequestURL && m_pNodeName)
		{
			if(m_pFetchURL)
				InsertNavigationBar(m_pRequestURL + strlen(m_pNodeName) + strlen(m_pFetchURL) + 1);
			else
				InsertNavigationBar(m_pRequestURL);
		}
	}
	if(sErrFilename)
		delete sErrFilename;

	if(sOut)
		delete sOut;

	return -1;
}

int PsiphonHandler::GetClientResponseContent(MaClientEx* pClient, char*& cont_type, char*& content, int& contentLen)
{
	content = pClient->getResponseContent(&contentLen);

	if(content == NULL)
		return RequestError(400);

	m_currentUser.SetRead(m_currentUser.GetRead() + contentLen);

	cont_type = pClient->getHeaderVar("content-type");

	if(cont_type == NULL)
		return RequestError(400);

	return 0;
}

int PsiphonHandler::ServeRequest()
{
	int nResult = ValidateRequestURI();

	SetSessionData();
	SetClientShowImages();


	switch(nResult)
	{
	case SERVEFILE:
		ServeFile();
		return 0;
	case LOGIN:
		Login();
		return 0;
	case LOGIN_REDIRECT:
		RedirectSpecial(URL_LOGIN);
		return 0;
	case CHANGEPWD:
		ResetPassword();
		return 0;
	case CHANGEPWD_REDIRECT:
		RedirectSpecial(URL_CHANGEPWD);
		return 0;
	case GETBOOKMARKS:
		SendBookmarks();
		return 0;
	case ADDBOOKMARK:
		AddBookmark();
		return 0;
	case DELETEBOOKMARK:
		DeleteBookmark();
		return 0;
	case MAINPAGE:
		ShowMainPage();
		return 0;
	case FETCH:
		free(m_pFetchURL);
		m_pFetchURL = strdup(URL_FETCH);
		m_bTopBar = true;
		Fetch();
		return 0;
	case FETCH_REDIRECT:
		RedirectSpecial(URL_FETCH);
		return 0;
	case FETCHI:
		free(m_pFetchURL);
		m_pFetchURL = strdup(URL_FETCHI);
		m_bTopBar = false;
		Fetch();
		return 0;
	case FETCHI_REDIRECT:
		RedirectSpecial(URL_FETCHI);
		return 0;
	case TOPBARFRAME:
		SendTopBar();
		return 0;
	case SETLANG:
		SetLanguageCookie();
		return 0;
	case INDEX:
		ShowIndexPage();
		return 0;
	case NOTVALID:
		return RequestError(404);
	}
	return RequestError(404);
}

void PsiphonHandler::ShowMainPage()
{
	if(!IsLogged()) //check if user is logged
	{
		RedirectSpecial(URL_LOGIN);	
		return;
	}
	m_pRq->setResponseCode(200);
	m_pRq->setHeader("Content-type: text/html", false);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	CPsiphonUtils util(this);
	char* template_buffer = NULL;
	util.GetHTML("main.tpl.html", template_buffer);

	if(template_buffer)
	{
		Write(template_buffer);
		delete template_buffer;
		InsertNavigationBar("");
	}
}

int PsiphonHandler::Fetch()
{
	if(!IsLogged()) //check if user is logged
	{
		char urlBuf[MPR_HTTP_MAX_URL];

		int nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "%s%s", URL_LOGIN, m_pRequestURL + strlen(m_pFetchURL) + strlen(m_pNodeName) + 1);
		if(nLen <= 0 || nLen >= MPR_HTTP_MAX_URL)
			return -1;

		m_pRq->setResponseCode(200);
		RedirectSpecial(urlBuf);

		return -1;
	}

	char* request_uri = strdup(m_pRequestURL + (strlen(m_pFetchURL) + strlen(m_pNodeName) + 1));

	//assume "http://" if no protocol entered
	if(!strstr(request_uri, "://"))
	{
		char* urlBuf = new char[MPR_HTTP_MAX_URL];	
		int nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "http://%s", request_uri);
		if(nLen <= 0 || nLen >= MPR_HTTP_MAX_URL)
		{
			delete urlBuf;
			return -1;
		}


		delete request_uri;	
		request_uri = urlBuf;
	}

	//validate protocol
	if(ValidateRequestProtocol(request_uri) != 0)
	{
		delete request_uri;	
		return -1;
	}

	char* http_method = NULL;
	if(ValidateRequest(http_method) != 0)
		return -1;

	if(!IsExternalHost(request_uri))
	{
		delete request_uri;	
		return RequestError(403);
	}

	MaUrl maUrl;
	maUrl.parse(request_uri);

	if(RedirectRequestURI(maUrl, request_uri) != 0)
	{
		delete request_uri;	
		return -1;
	}

	MprBuf bufHeader;
	PrepareHeader(bufHeader, http_method, request_uri, maUrl);

	unsigned long tStart = GetTime();

	MaClientEx* pClient = new MaClientEx;

	pClient->allowContentType("text/");
	pClient->allowContentType("application/msword");
	pClient->allowContentType("application/pdf");
	pClient->allowContentType("application/postscript");
	pClient->allowContentType("application/xhtml+xml");

	if(m_bServerShowImages && m_bClientShowImages)
	{
		pClient->allowContentType("image/bmp");
		pClient->allowContentType("image/gif");
		pClient->allowContentType("image/ief");
		pClient->allowContentType("image/jpeg");
		pClient->allowContentType("image/png");
		pClient->allowContentType("image/tiff");
		pClient->allowContentType("image/x-icon");
	}

	int rc = pClient->sendRequest(maUrl.host, maUrl.port, &bufHeader, m_pPostBuf->getStart(), m_pPostBuf->getLength());

	unsigned long tEnd = GetTime();
	unsigned long nReadTime = tEnd - tStart;
	m_currentUser.SetReadTime(m_currentUser.GetReadTime() + nReadTime);

	if(rc != 0)
	{
		switch(rc)
		{
		case MPR_ERR_TIMEOUT:
			RequestError(504);
			break;
		case MPR_ERR_BAD_TYPE:
			RequestError(403);
			break;
		default:
			RequestError(400);
			break;
		}
		delete pClient;	
		delete request_uri;	
		return -1;
	}

	//	Get the response header and examine for cookies before redirects(sic!)
	char* response_header = pClient->getResponseHeader();
	if(response_header == NULL)
	{
		RequestError(400);
		delete pClient;	
		delete request_uri;	
		return -1;
	}

	StoreCookies(response_header, request_uri);

	int nCode = pClient->getResponseCode();
	if(ExamineResponseCode(nCode, pClient, request_uri) != 0)
	{
		delete pClient;	
		delete request_uri;	
		return -1;
	}

	//	Get the actual response content
	int contentLen = 0;
	char* content = NULL;
	char* cont_type = NULL;

	if(GetClientResponseContent(pClient, cont_type, content, contentLen) != 0)
	{
		delete pClient;	
		delete request_uri;	
		return -1;
	}

	WriteResponse(cont_type, content, contentLen, request_uri);

	delete request_uri;	
	delete pClient;
	return 0;
}

void PsiphonHandler::RedirectSpecial(char* sUrl)
{
	if(!m_pNodeName)
		return;

	if(!sUrl)
		return;

	if((strlen(m_pNodeName) + strlen(sUrl)) > MPR_HTTP_MAX_URL - 1)
		return;

	char urlBuf[MPR_HTTP_MAX_URL];	

	int nLen = _snprintf(urlBuf, MPR_HTTP_MAX_URL, "/%s%s", m_pNodeName, sUrl);
	if(nLen <= 0 || nLen >= MPR_HTTP_MAX_URL)
		return;

	char* redirect_path = urlBuf;
	m_pRq->redirect(302, redirect_path);
}

void PsiphonHandler::SendTopBar()
{
	if(!IsLogged()) //check if user is logged
	{
		RequestError(404);	
		return;
	}
	m_pRq->setResponseCode(200);
	m_pRq->setHeaderFlags(MPR_HTTP_DONT_CACHE, 0);
	m_pRq->setHeader("Content-type: text/html", true);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	Write("<HTML><BODY>");

	InsertNavigationBar(m_pRequestURL + strlen("/topbar-frame/"));	

	Write("</BODY></HTML>");
	return;
}

void PsiphonHandler::SetLanguageCookie()
{
	char* language = m_pRq->getVar(MA_FORM_OBJ, "lang", "");
	char* redirect_url = m_pRq->getVar(MA_FORM_OBJ, "url", "");

	m_pRq->setCookie("psi_lang", language, 0, "/", true); 
	m_pRq->setResponseCode(200);
	m_pRq->redirect(302, redirect_url);
	return;
}

int PsiphonHandler::WriteResponse(char* cont_type, char* content, int contentLen, char* request_uri)
{
	char headerBuf[MPR_HTTP_MAX_HEADER];	
	int nLen = _snprintf(headerBuf, MPR_HTTP_MAX_HEADER, "Content-type: %s", cont_type);
	if(nLen <= 0 || nLen >= MPR_HTTP_MAX_HEADER)
		return -1;

	m_pRq->setResponseCode(200);
	m_pRq->setHeaderFlags(MPR_HTTP_DONT_CACHE, 0);
	m_pRq->setHeader(headerBuf, true);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	int n_filter = NO_FILTER;

	if(StartsWith(cont_type, "text/html"))
		n_filter = HTML_FILTER;
	else if(StartsWith(cont_type, "text/css"))
		n_filter = CSS_FILTER;

	FilterWrite(n_filter, content, contentLen, request_uri);

	return 0;
}

void PsiphonHandler::FilterWrite(int n_filter, char* content, int contentLen, char* request_uri)
{
	if(n_filter == NO_FILTER)
	{
		Write(content, contentLen);
		return;
	}

	bool stripImages = !(m_bServerShowImages && m_bClientShowImages);

	CPsiphonContFilter* filter = NULL;
	
	filter = new CPsiphonContFilter(this, true, true, stripImages);

	std::istrstream* sStream = new std::istrstream(content);

	if(n_filter == HTML_FILTER)
	{
		filter->ProcessContent(sStream, request_uri);

		if(m_bTopBar)
			InsertNavigationBar(request_uri);
	}
	else if(n_filter == CSS_FILTER)
	{
		filter->ProcessCSS(sStream, request_uri);
	}

	if(sStream)
		delete sStream;
	if(filter)
		delete filter;
}

void PsiphonHandler::SetSessionData()
{
	CPsiphonUtils util(this);
	const char* pSessVal = util.GetCookieData("psi_sess");

	if(pSessVal && *pSessVal)
	{
		int nUserID = 0;

		CDBSessions dbSessions;
		int rc = dbSessions.FindDBSession(pSessVal, nUserID); 

		if(rc == 0)
			m_currentUser.SetID(nUserID);
	}

	free((void*)pSessVal);
}

int PsiphonHandler::IsLogged()
{
	if(m_currentUser.GetID())
		return 1;

	return 0;
}

void UserLog(const char* msg)
{
	char tmp_msg[255];
	int nLen = _snprintf(tmp_msg, 255, "%s", msg);
	if(nLen <= 0 || nLen >= 255)
		return;

	mprLog("%s", (char*)tmp_msg);
}

void PsiphonHandler::ResetPassword()
{
	int n_error = 0;
	char* user = m_pRq->getVar(MA_FORM_OBJ, "user", "");
	char* pwd = m_pRq->getVar(MA_FORM_OBJ, "password", "");
	char* new_pwd = m_pRq->getVar(MA_FORM_OBJ, "new_password", "");
	char* confirm_pwd = m_pRq->getVar(MA_FORM_OBJ, "confirm_password", "");
	char* tpl_filename = "reset_pwd.tpl.html"; //Change password

	if(strlen(m_pRq->getVar(MA_FORM_OBJ, "Submit", "")))
	{
		if(!strlen(user))
		{
			n_error = 1;
			tpl_filename = "reset_pwd_err_user.tpl.html";// "Enter your username!";
		}
		else if(!strlen(pwd))
		{
			n_error = 1;
			tpl_filename = "reset_pwd_err_cur_pwd.tpl.html";// "Enter your current password!";
		}
		else if(!strlen(new_pwd))
		{
			n_error = 1;
			tpl_filename = "reset_pwd_err_new_pwd.tpl.html";// "Enter your new password!";
		}
		else if(!strlen(confirm_pwd))
		{
			n_error = 1;
			tpl_filename = "reset_pwd_err_conf_pwd.tpl.html";// "Confirm your new password!";
		}
		else if(strcmp(new_pwd, confirm_pwd))
		{
			n_error = 1;
			tpl_filename = "reset_pwd_err_conf_fail_pwd.tpl.html";// "Password confirmation failed!";
		}
	}
	else
		n_error = 1;

	m_pRq->setResponseCode(200);
	m_pRq->setHeaderFlags(MPR_HTTP_DONT_CACHE, 0);

	int nUserID = 0;
	char pwd_encrypted[64];
	if(!n_error && strlen(user) && strlen(pwd))
	{
		CDBUser::EncryptPassword(pwd_encrypted, 64, user, pwd);
		nUserID = CDBUser::FindUser(user, pwd_encrypted);
	}

	if(!n_error && nUserID <= 0)
	{
		n_error = 1;
		tpl_filename = "reset_pwd_err_no_match.tpl.html";// "Incorrect name/password!";
	}

	if(!n_error)
	{
		memset(pwd_encrypted, 0, 64);

		CDBUser::EncryptPassword(pwd_encrypted, 64, user, new_pwd);

		if(CDBUser::ResetPassword(nUserID, pwd_encrypted) != 0)
		{
			n_error = 1;
			tpl_filename = "reset_pwd_err_db_update.tpl.html";// "DB Error: can't reset password";
		}
	}

	m_pRq->setHeader("Content-type: text/html", false);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	CPsiphonUtils util(this);

	if(n_error)
	{
		char* template_buffer = NULL;
		util.GetHTML(tpl_filename, template_buffer);

		if(!template_buffer)
			return;
		std::vector <CStringReplaceData> replace_ar;
		replace_ar.reserve(2);
		replace_ar.push_back(CStringReplaceData("username", user));
		replace_ar.push_back(CStringReplaceData("nodename", m_pNodeName));

		ReplaceWrite(template_buffer, replace_ar);		

		delete template_buffer;
	}
	else
	{
		char* template_buffer = NULL;
		util.GetHTML("reset_pwd_success.tpl.html", template_buffer);

		if(!template_buffer)
			return;

		std::vector <CStringReplaceData> replace_ar;
		replace_ar.reserve(1);
		replace_ar.push_back(CStringReplaceData("nodename", m_pNodeName));

		ReplaceWrite(template_buffer, replace_ar);		
		delete template_buffer;
	}
}

void PsiphonHandler::SendBookmarks()
{
	m_pRq->setResponseCode(200);
	m_pRq->setHeaderFlags(MPR_HTTP_DONT_CACHE, 0);
	m_pRq->setHeader("Content-type: text/html", false);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	CDBUserBookmarks dbBookmarks;
	char* sBookMarks = dbBookmarks.BuildDBBookMarks(m_currentUser.GetID());

	if(!sBookMarks)
		return;

	Write(sBookMarks);
	delete sBookMarks;
}

void PsiphonHandler::AddBookmark()
{
	char* sBookMark = m_pRq->getVar(MA_FORM_OBJ, "bookmark", NULL);
	if(sBookMark == NULL || *sBookMark == 0)
		return;

	CDBUserBookmarks dbBookmarks;
	int nRes = dbBookmarks.AddDBBookMark(sBookMark, m_currentUser.GetID());

	m_pRq->setResponseCode(200);
	m_pRq->setHeaderFlags(MPR_HTTP_DONT_CACHE, 0);
	m_pRq->setHeader("Content-type: text/html", false);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	if(nRes == 0)
		Write("OK\n");
	else
		Write("ERROR\n");

	Write(sBookMark);
}

void PsiphonHandler::DeleteBookmark()
{
	char* sBookMark = m_pRq->getVar(MA_FORM_OBJ, "bookmark", NULL);

	if(sBookMark == NULL || *sBookMark == 0)
		return;

	CDBUserBookmarks dbBookmarks;
	int nRes = dbBookmarks.DeleteDBBookMark(sBookMark, m_currentUser.GetID());

	m_pRq->setResponseCode(200);
	m_pRq->setHeaderFlags(MPR_HTTP_DONT_CACHE, 0);
	m_pRq->setHeader("Content-type: text/html", false);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	if(nRes == 0)
		Write("OK\n");
	else
		Write("ERROR\n");

	Write(sBookMark);
}

void PsiphonHandler::Login()
{
	char* tpl_filename = NULL;
	char* user = NULL;
	char* request_uri = m_pRequestURL + strlen(m_pNodeName) + strlen(URL_LOGIN) + 1; 

	char redirectBuf[MPR_HTTP_MAX_URL];	

	int nLen = 0;
	if(request_uri && *request_uri)
		nLen = _snprintf(redirectBuf, MPR_HTTP_MAX_URL, "%s%s", URL_FETCH, request_uri);
	else
		nLen = _snprintf(redirectBuf, MPR_HTTP_MAX_URL, "%s", URL_FETCH);

	if(nLen <= 0 || nLen >= MPR_HTTP_MAX_URL)
		return;

	if(!TryToLogin(tpl_filename, user))
		ShowLoginPage(tpl_filename, user);
	else
		RedirectSpecial(redirectBuf);
}

bool PsiphonHandler::TryToLogin(char*& filename, char*& user)
{
	CDBUser::SetLogFunc(UserLog);

	m_pRq->setCookie("psi_sess", "", 0, "/", true); 

	user = m_pRq->getVar(MA_FORM_OBJ, "user", "");
	char* pwd = m_pRq->getVar(MA_FORM_OBJ, "password", "");

	filename = "login.tpl.html";

	if(!strlen(m_pRq->getVar(MA_FORM_OBJ, "Submit", "")))
		return false;

	if(!strlen(user))
	{
		filename = "login_err_user.tpl.html";
		return false;
	}
	else if(!strlen(pwd))
	{
		filename = "login_err_pwd.tpl.html";
		return false;
	}

	m_pRq->setResponseCode(200);
	m_pRq->setHeaderFlags(MPR_HTTP_DONT_CACHE, 0);

	char pwd_encrypted[64];
	CDBUser::EncryptPassword(pwd_encrypted, 64, user, pwd);

	int nUserID = CDBUser::FindUser(user, pwd_encrypted);

	if(nUserID <= 0)
	{
		filename = "login_err_no_match.tpl.html";
		return false;
	}

	CDBUserCookies dbCookies;
	dbCookies.PurgeDBUserCookies(nUserID);

	CDBSessions dbSessions;

	char seed[255];
	int nLen = _snprintf(seed, 255, "%d", GetTime());
	if(nLen <= 0 || nLen >= 255)
	{
		filename = "login_err_db_insert_sess.tpl.html";
		return false;
	}

	CSHA1 sha1;
	sha1.Update((UINT_8*)seed, strlen(seed));
	sha1.Update((UINT_8*)pwd, strlen(pwd));
	sha1.Update((UINT_8*)user, strlen(user));
	sha1.Final();

	char sSessVal[64];
	sSessVal[0] = 0;
	sha1.ReportHash(sSessVal, CSHA1::REPORT_HEX);

	if(!dbSessions.CreateDBSession(nUserID, sSessVal))
	{
		filename = "login_err_db_insert_sess.tpl.html";
		return false;
	}

	if(sSessVal == NULL || *sSessVal == '\0')
		return false;

	m_pRq->setCookie("psi_sess", sSessVal, 0, "/", true); 

	return true;
}

void PsiphonHandler::ShowLoginPage(char* filename, char* user)
{
	char* pLanguages = NULL;

	char* template_buffer = NULL;

	CPsiphonUtils util(this);
	util.GetHTML(filename, template_buffer);

	if(!template_buffer)
		return;

	CLanguageData lang[MAX_LANG_SIZE];
	util.GetLanguagesData(lang, MAX_LANG_SIZE);

	GetLanguagesStr(lang, &pLanguages);

	if(!pLanguages)
		pLanguages = "";

	m_pRq->setResponseCode(200);
	m_pRq->setHeader("Content-type: text/html", false);
	m_pRq->insertDataStream(m_pRq->getDynBuf());

	std::vector <CStringReplaceData> replace_ar;
	replace_ar.reserve(3);
	replace_ar.push_back(CStringReplaceData("username", user));
	replace_ar.push_back(CStringReplaceData("nodename", m_pNodeName));
	replace_ar.push_back(CStringReplaceData("languages", pLanguages));

	ReplaceWrite(template_buffer, replace_ar);		

	delete pLanguages;
	delete template_buffer;
}

int PsiphonHandler::InsertNavigationBar(char* request_uri)
{
	char* template_buffer = NULL;

	CPsiphonUtils util(this);
	util.GetHTML("topbar.tpl.html", template_buffer);

	if(!template_buffer)
		return 0;

	char* sImagesOn = "0";

	if(m_bServerShowImages)
		sImagesOn = "1";

	
	std::vector <CStringReplaceData> replace_ar;
	replace_ar.reserve(3);
	replace_ar.push_back(CStringReplaceData("images_on", sImagesOn));
	replace_ar.push_back(CStringReplaceData("nodename", m_pNodeName));
	replace_ar.push_back(CStringReplaceData("url", request_uri));

	ReplaceWrite(template_buffer, replace_ar);		

	delete template_buffer;
	return 0;
}

int PsiphonHandler::Write(const char* _pBuf, int nSize)
{
	int nWritten = 0;

	unsigned long tStart = GetTime();

	if(nSize == -1)
		nWritten = m_pRq->write((char*)_pBuf);
	else
		nWritten = m_pRq->write((char*)_pBuf, nSize);

	unsigned long tEnd = GetTime();
	unsigned long tWriteTime = tEnd - tStart;

	m_currentUser.SetWriteTime(m_currentUser.GetWriteTime() + tWriteTime);
	m_currentUser.SetWritten(m_currentUser.GetWritten() + nWritten);

	return nWritten;
}

int PsiphonHandler::ReplaceWrite(char* buf, std::vector <CStringReplaceData> replaceArr)
{
	char* start_tag;
	char* offset;
	int nWritten = 0;
	bool bFound = false;
	
	if(replaceArr.empty())
	{
		nWritten = Write(buf);
		return nWritten;
	}

	//scan buf for replacement tag '<%'
	while((start_tag = strstr(buf, "<%")) != NULL)
	{
		//reset previous results
		offset = NULL;
		bFound = false;

		nWritten += Write(buf, start_tag - buf);

		//replacement tag found, offset to the replace var name
		offset = start_tag + 2;

		for(int i = 0; i < (int)replaceArr.size(); i++)
		{
			if(replaceArr.at(i).m_varname == NULL)
				continue;
		
			// see if current element matches replace var name beginning
			if(strncmp(offset, replaceArr.at(i).m_varname, strlen(replaceArr.at(i).m_varname)) != 0)
				continue; //try next element in the array

			//match found, offset to the suspected var end 
			int nLen = strlen(replaceArr.at(i).m_varname);
			offset += nLen;

			//check if offset points to replacement end tag '%>'
			if(strncmp(offset, "%>", 2) != 0)
			{
				offset -= nLen; //move offset back			
				continue; //try next element in the array
			}

			offset += 2;

			bFound = true;
			nWritten += Write(replaceArr.at(i).m_replacement);
			break;
		}

		if(!bFound)
			nWritten += Write("<%");

		buf = offset;
	}

	nWritten += Write(buf);

	return nWritten;
}


const char* PsiphonHandler::GetCleanURI(const char* pURI)
{
	const char* pCleanURI = pURI;
	char* header = "/";
	
	if(!StartsWith(pCleanURI, header))
		return pURI;

	pCleanURI += strlen(header);
//
	header = m_pNodeName;

	if(!StartsWith(pCleanURI, header))
		return pURI;

	pCleanURI += strlen(header);
//
	header = "/";

	if(!StartsWith(pCleanURI, header))
		return pURI;

	pCleanURI += strlen(header);
//

	header = "fetch/";

	if(StartsWith(pCleanURI, header))
	{
		pCleanURI += strlen(header);
		return pCleanURI;
	}

	header = "fetch-i/";

	if(StartsWith(pCleanURI, header))
	{
		pCleanURI += strlen(header);
		return pCleanURI;
	}

	return pURI;
}

void PsiphonHandler::LogRequest()
{
	if(m_currentUser.GetID() <= 0)
		return;

	if(m_currentUser.GetWritten() == 0 && m_currentUser.GetRead() == 0)
		return;

	if(m_currentUser.GetWriteTime() == 0 && m_currentUser.GetReadTime() == 0)
		return;

	CPsiphonMessage log;
	
	log.SetMessage(PSIPHON_MESSAGE_REQUEST);
	log.SetIP(m_pRq->getVar(MA_REQUEST_OBJ, "REMOTE_HOST", 0));
	
	log.SetURI(GetCleanURI(m_pRequestURL));
	log.SetUserID(m_currentUser.GetID());
	log.SetLastAccess((int)time(NULL));
	log.SetWritten(m_currentUser.GetWritten());
	log.SetRead(m_currentUser.GetRead());
	log.SetWriteTime(m_currentUser.GetWriteTime());
	log.SetReadTime(m_currentUser.GetReadTime());

	char* pOut = NULL;
	
	log.Out(pOut);
	if(pOut == NULL)
		return;
    
	mprLog("%s", (char*)pOut);
}

int PsiphonHandler::GetLanguagesStr(CLanguageData lang[], char** pLanguages)
{
	char* s_original_url = m_pRq->getOriginalUri();

	char* s_urlencoded = new char[strlen(s_original_url) * 3 + 1];
	maUrlEncode(s_urlencoded, strlen(s_original_url) * 3 + 1, s_original_url, true);

	char* fmt = "%s <a href=\"/%s/set-lang/?lang=%s&url=%s\">%s</a>";

	for(int i = 0;i < MAX_LANG_SIZE;i++)
	{
		char* s_lang_dir = lang[i].GetDir();
		char* s_language = lang[i].GetLanguage();

		if(!s_lang_dir || !s_language)
			break;

		char* strBuf = NULL;

		if(*pLanguages && **pLanguages)
		{
			int nSize = strlen(m_pNodeName) + strlen(*pLanguages) + strlen(s_urlencoded) + strlen(fmt) + strlen(s_lang_dir) + strlen(s_language) + 1;		

			strBuf = new char[nSize]; 
			int nLen = _snprintf(strBuf, nSize, fmt, *pLanguages, m_pNodeName, s_lang_dir, s_urlencoded, s_language);
			if(nLen <= 0 || nLen >= nSize)
			{
				delete strBuf;
				return -1;
			}
		}
		else
		{
			int nSize = strlen(m_pNodeName) + strlen(s_urlencoded) + strlen(fmt) + strlen(s_lang_dir) + strlen(s_language) + 1;

			strBuf = new char[nSize]; 
			int nLen = _snprintf(strBuf, nSize, fmt, "", m_pNodeName, s_lang_dir, s_urlencoded, s_language);
			if(nLen <= 0 || nLen >= nSize)
			{
				delete strBuf;
				return -1;
			}
		}

		delete * pLanguages;
		*pLanguages = strBuf;
	}

	delete(s_urlencoded);

	return 0;
}

void PsiphonHandler::ServeFile()
{
	if(!IsLogged())
	{
		RequestError(404);
		return;
	}
	char* fileName = NULL;
	int flags = 0;

	struct stat fileinfo;

	fileName = m_pRequestURL + strlen(m_pNodeName) + 1 + strlen(URL_SERVEFILE);

	char* docRoot = m_pRq->host->getDocumentRoot();

	int nSize = strlen(docRoot) + strlen(URL_SERVEFILE) + strlen(fileName) + 1;

	char* path = new char[nSize];
	int nLen = _snprintf(path, nSize, "%s%s%s", docRoot, URL_SERVEFILE, fileName);
	if(nLen <= 0 || nLen >= nSize)
	{
		delete path;
		RequestError(404);
		return;
	}

	if(stat(path, &fileinfo) != 0)
	{
		delete path;
		RequestError(404);
		return;
	}

	flags = m_pRq->getFlags();

	if(m_pRq->openDoc(path) < 0)
	{
		delete path;
		RequestError(404);
		return;
	} 
	if(flags & (MPR_HTTP_GET_REQUEST | MPR_HTTP_HEAD_REQUEST | MPR_HTTP_POST_REQUEST))
	{
		m_pRq->insertDataStream(m_pRq->getDocBuf());
		m_pRq->getDocBuf()->setSize(fileinfo.st_size);
	}

	delete path;
	m_pRq->setResponseCode(200);
}

bool PsiphonHandler::IsExternalHost(char* pUri)
{
	MaUrl maUrl;
	maUrl.parse(pUri);

	if(IsNumericAddress(maUrl.host))
	{
		return CPsiphonUtils::IsExternalIP(maUrl.host);
	}
	else
	{
		struct hostent* hp;
		char addrStr[MPR_MAX_IP_ADDR];

		hp = mprGetHostByName(maUrl.host);
		if(hp != 0)
		{
			for(int i = 0;hp->h_addr_list[i];i++)
			{
				mprInetNtoa(addrStr, sizeof(addrStr), *((struct in_addr *)hp->h_addr_list[i]));
				if(!CPsiphonUtils::IsExternalIP(addrStr))
					return false;
			}
		}
	}

	return true;
}

void PsiphonHandler::SetClientShowImages()
{
	
	m_bClientShowImages = true;

	CPsiphonUtils util(this);
	const char* pClientShowImages = util.GetCookieData("images");

	if(pClientShowImages)
	{
		m_bClientShowImages = atoi(pClientShowImages);
		free((void*)pClientShowImages);
	}

}

